#include "fsck.h"

uint64_t n_groups;
uint64_t inodes_in_last_group;
uint64_t inode_bitmap_inodes;

unsigned log2_pages_per_group;
unsigned log2_inodes_per_group;

unsigned log2_bytes_per_group;

uint64_t max_bytes;

static struct nvfs_sb *sb2 = NULL;

void superblock_modified(void)
{
	uint8_t checksum;
	unsigned i;
	sb->checksum = 0;
	checksum = 0;
	for (i = 0; i < NVFS_SB_STATIC_PART; i++)
		checksum += ((uint8_t *)sb)[i];
	sb->checksum = -checksum;
	memcpy(sb2, sb, NVFS_SB_STATIC_PART);
}

static int find_backup_superblock(void)
{
	int l2ps;
	for (l2ps = MIN_LOG_PAGE_SIZE; l2ps <= MAX_LOG_PAGE_SIZE; l2ps++) {
		int ps;
		size_t pos;
		struct nvfs_sb *sbb;
		ps = 1 << l2ps;
		pos = size & -ps;
		if (!pos)
			continue;
		sbb = ptr + (pos - ps);
		if (!(nvfs_validate_super(sbb, size))) {
			static struct qkey q;
			if (fsck_query(&q, Q_NOAUTO, "backup superblock found", "restore?")) {
				memcpy(sb, sbb, NVFS_SB_STATIC_PART);
				memset((void *)sb + NVFS_SB_STATIC_PART, 0, NVFS_SB_DYNAMIC_PART);
				sb->dirty = cpu_to_le32(1);
				return 0;
			} else {
				return 4;
			}
		}
	}
	return 4;
}

int check_super(void)
{
	const char *vs;
	bool ovw;

	fsck_log("checking super block...");

	vs = nvfs_validate_super(ptr, size);
	if (vs) {
		int r;
		static struct qkey q;
		fsck_query(&q, Q_JUST_PRINT, "%s", NULL, vs);
		r = find_backup_superblock();
		if (r)
			return r;
	}

	if (size < (le64_to_cpu(sb->n_pages) << sb->log2_page_size) + NVFS_SB_STATIC_PART)
		goto ret_0;
	sb2 = (void *)ptr + (le64_to_cpu(sb->n_pages) << sb->log2_page_size);

	ovw = false;
	if ((vs = nvfs_validate_super(sb2, size))) {
		static struct qkey q;
		ovw = fsck_query(&q, Q_NOAUTO, "invalid backup superblock: %s", "overwrite?", vs);
		force = 1;
	} else if (memcmp(sb, sb2, NVFS_SB_STATIC_PART)) {
		static struct qkey q;
		ovw = fsck_query(&q, Q_NOAUTO, "the backup superblock doesn't match the primary superblock", "overwrite?");
		force = 1;
	}
	if (ovw)
		memcpy(sb2, sb, NVFS_SB_STATIC_PART);

ret_0:
	nvfs_get_sizes(sb, &n_groups, NULL, NULL, &inodes_in_last_group, &inode_bitmap_inodes, &max_bytes);
	log2_pages_per_group = sb->log2_page_size + 3;
	log2_inodes_per_group = sb->log2_page_size - sb->log2_inode_size + sb->log2_inode_pages_per_group;
	log2_bytes_per_group = sb->log2_page_size * 2 + 3;

	if (le32_to_cpu(sb->error)) {
		static struct qkey q;
		fsck_query(&q, Q_JUST_PRINT, "the filesystem has the structure error flag.", NULL);
	}
	return 0;
}
